@@ -118,14 +118,18 @@ _downcast_and_check(Py_ssize_t size, int extra) {
  */
 int convert_codec_options(PyObject* options_obj, void* p) {
     codec_options_t* options = (codec_options_t*)p;
-    if (!PyArg_ParseTuple(options_obj, "Obb",
+    options->unicode_decode_error_handler = NULL;
+    if (!PyArg_ParseTuple(options_obj, "ObbzO",
                           &options->document_class,
                           &options->tz_aware,
-                          &options->uuid_rep)) {
+                          &options->uuid_rep,
+                          &options->unicode_decode_error_handler,
+                          &options->tzinfo)) {
         return 0;
     }
 
     Py_INCREF(options->document_class);
+    Py_INCREF(options->tzinfo);
     return 1;
 }
 
@@ -137,10 +141,14 @@ void default_codec_options(codec_options_t* options) {
     // TODO: set to "1". PYTHON-526, setting tz_aware=True by default.
     options->tz_aware = 0;
     options->uuid_rep = PYTHON_LEGACY;
+    options->unicode_decode_error_handler = NULL;
+    options->tzinfo = Py_None;
+    Py_INCREF(options->tzinfo);
 }
 
 void destroy_codec_options(codec_options_t* options) {
     Py_CLEAR(options->document_class);
+    Py_CLEAR(options->tzinfo);
 }
 
 static PyObject* elements_to_dict(PyObject* self, const char* string,
@@ -1560,7 +1568,9 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             if (buffer[*position + value_length - 1]) {
                 goto invalid;
             }
-            value = PyUnicode_DecodeUTF8(buffer + *position, value_length - 1, "strict");
+            value = PyUnicode_DecodeUTF8(
+                buffer + *position, value_length - 1,
+                options->unicode_decode_error_handler);
             if (!value) {
                 goto invalid;
             }
@@ -1687,7 +1697,11 @@ static PyObject* get_value(PyObject* self, const char* buffer,
                     Py_DECREF(value);
                     goto invalid;
                 }
-                PyList_Append(value, to_append);
+                if (PyList_Append(value, to_append) < 0) {
+                    Py_DECREF(value);
+                    Py_DECREF(to_append);
+                    goto invalid;
+                }
                 Py_DECREF(to_append);
             }
             (*position)++;
@@ -1860,6 +1874,7 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             PyObject* replace;
             PyObject* args;
             PyObject* kwargs;
+            PyObject* astimezone;
             long long millis;
             if (max < 8) {
                 goto invalid;
@@ -1901,6 +1916,26 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             }
             Py_XDECREF(utc_type);
             value = PyObject_Call(replace, args, kwargs);
+            if (!value) {
+                Py_DECREF(replace);
+                Py_DECREF(args);
+                Py_DECREF(kwargs);
+                goto invalid;
+            }
+
+            /* convert to local time */
+            if (options->tzinfo != Py_None) {
+                astimezone = PyObject_GetAttrString(value, "astimezone");
+                if (!astimezone) {
+                    Py_DECREF(replace);
+                    Py_DECREF(args);
+                    Py_DECREF(kwargs);
+                    goto invalid;
+                }
+                value = PyObject_CallFunctionObjArgs(astimezone, options->tzinfo, NULL);
+                Py_DECREF(astimezone);
+            }
+
             Py_DECREF(replace);
             Py_DECREF(args);
             Py_DECREF(kwargs);
@@ -1916,7 +1951,9 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             if (pattern_length > BSON_MAX_SIZE || max < pattern_length) {
                 goto invalid;
             }
-            pattern = PyUnicode_DecodeUTF8(buffer + *position, pattern_length, "strict");
+            pattern = PyUnicode_DecodeUTF8(
+                buffer + *position, pattern_length,
+                options->unicode_decode_error_handler);
             if (!pattern) {
                 goto invalid;
             }
@@ -1980,8 +2017,9 @@ static PyObject* get_value(PyObject* self, const char* buffer,
                 goto invalid;
             }
 
-            collection = PyUnicode_DecodeUTF8(buffer + *position,
-                                              coll_length - 1, "strict");
+            collection = PyUnicode_DecodeUTF8(
+                buffer + *position, coll_length - 1,
+                options->unicode_decode_error_handler);
             if (!collection) {
                 goto invalid;
             }
@@ -2026,7 +2064,9 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             if (buffer[*position + value_length - 1]) {
                 goto invalid;
             }
-            code = PyUnicode_DecodeUTF8(buffer + *position, value_length - 1, "strict");
+            code = PyUnicode_DecodeUTF8(
+                buffer + *position, value_length - 1,
+                options->unicode_decode_error_handler);
             if (!code) {
                 goto invalid;
             }
@@ -2068,7 +2108,9 @@ static PyObject* get_value(PyObject* self, const char* buffer,
             if (buffer[*position + code_size - 1]) {
                 goto invalid;
             }
-            code = PyUnicode_DecodeUTF8(buffer + *position, code_size - 1, "strict");
+            code = PyUnicode_DecodeUTF8(
+                buffer + *position, code_size - 1,
+                options->unicode_decode_error_handler);
             if (!code) {
                 goto invalid;
             }
@@ -2261,8 +2303,29 @@ static PyObject* _elements_to_dict(PyObject* self, const char* string,
             Py_DECREF(dict);
             return NULL;
         }
-        name = PyUnicode_DecodeUTF8(string + position, name_length, "strict");
+        name = PyUnicode_DecodeUTF8(
+            string + position, name_length,
+            options->unicode_decode_error_handler);
         if (!name) {
+            /* If NULL is returned then wrap the UnicodeDecodeError
+               in an InvalidBSON error */
+            PyObject *etype, *evalue, *etrace;
+            PyObject *InvalidBSON;
+
+            PyErr_Fetch(&etype, &evalue, &etrace);
+            InvalidBSON = _error("InvalidBSON");
+            if (InvalidBSON) {
+                Py_DECREF(etype);
+                etype = InvalidBSON;
+
+                if (evalue) {
+                    PyObject *msg = PyObject_Str(evalue);
+                    Py_DECREF(evalue);
+                    evalue = msg;
+                }
+                PyErr_NormalizeException(&etype, &evalue, &etrace);
+            }
+            PyErr_Restore(etype, evalue, etrace);
             Py_DECREF(dict);
             return NULL;
         }
@@ -2475,7 +2538,12 @@ static PyObject* _cbson_decode_all(PyObject* self, PyObject* args) {
             destroy_codec_options(&options);
             return NULL;
         }
-        PyList_Append(result, dict);
+        if (PyList_Append(result, dict) < 0) {
+            Py_DECREF(dict);
+            Py_DECREF(result);
+            destroy_codec_options(&options);
+            return NULL;
+        }
         Py_DECREF(dict);
         string += size;
         total_size -= size;
