@@ -15,6 +15,7 @@
 """Collection level utilities for Mongo."""
 
 import collections
+import datetime
 import warnings
 
 from bson.code import Code
@@ -33,7 +34,6 @@ from pymongo.command_cursor import CommandCursor
 from pymongo.cursor import Cursor
 from pymongo.errors import ConfigurationError, InvalidName, OperationFailure
 from pymongo.helpers import _check_write_command_response
-from pymongo.message import _INSERT, _UPDATE, _DELETE
 from pymongo.operations import _WriteOp, IndexModel
 from pymongo.read_preferences import ReadPreference
 from pymongo.results import (BulkWriteResult,
@@ -364,6 +364,8 @@ class Collection(common.BaseObject):
         :Returns:
           An instance of :class:`~pymongo.results.BulkWriteResult`.
 
+        .. seealso:: :ref:`writes-and-ids`
+
         .. versionadded:: 3.0
         """
         if not isinstance(requests, list):
@@ -380,13 +382,91 @@ class Collection(common.BaseObject):
             return BulkWriteResult(bulk_api_result, True)
         return BulkWriteResult({}, False)
 
-    def _insert(self, sock_info, docs, ordered=True,
-                check_keys=True, manipulate=False, write_concern=None):
+    def _legacy_write(
+            self, sock_info, name, cmd, acknowledged, op_id, func, *args):
+        """Internal legacy write helper."""
+        listeners = self.database.client._event_listeners
+        publish = listeners.enabled_for_commands
+
+        if publish:
+            start = datetime.datetime.now()
+        rqst_id, msg, max_size = func(*args)
+        if publish:
+            duration = datetime.datetime.now() - start
+            listeners.publish_command_start(
+                cmd, self.__database.name, rqst_id, sock_info.address, op_id)
+            start = datetime.datetime.now()
+        try:
+            result = sock_info.legacy_write(
+                rqst_id, msg, max_size, acknowledged)
+        except Exception as exc:
+            if publish:
+                dur = (datetime.datetime.now() - start) + duration
+                if isinstance(exc, OperationFailure):
+                    details = exc.details
+                    # Succeed if GLE was successful and this is a write error.
+                    if details.get("ok") and "n" in details:
+                        reply = message._convert_write_result(
+                            name, cmd, details)
+                        listeners.publish_command_success(
+                            dur, reply, name, rqst_id, sock_info.address, op_id)
+                        raise
+                else:
+                    details = message._convert_exception(exc)
+                listeners.publish_command_failure(
+                    dur, details, name, rqst_id, sock_info.address, op_id)
+            raise
+        if publish:
+            if result is not None:
+                reply = message._convert_write_result(name, cmd, result)
+            else:
+                # Comply with APM spec.
+                reply = {'ok': 1}
+            duration = (datetime.datetime.now() - start) + duration
+            listeners.publish_command_success(
+                duration, reply, name, rqst_id, sock_info.address, op_id)
+        return result
+
+    def _insert_one(
+            self, sock_info, doc, ordered,
+            check_keys, manipulate, write_concern, op_id):
+        """Internal helper for inserting a single document."""
+        if manipulate:
+            doc = self.__database._apply_incoming_manipulators(doc, self)
+            if '_id' not in doc:
+                doc['_id'] = ObjectId()
+            doc = self.__database._apply_incoming_copying_manipulators(doc,
+                                                                       self)
+        concern = (write_concern or self.write_concern).document
+        acknowledged = concern.get("w") != 0
+        command = SON([('insert', self.name),
+                       ('ordered', ordered),
+                       ('documents', [doc])])
+        if concern:
+            command['writeConcern'] = concern
+
+        if sock_info.max_wire_version > 1 and acknowledged:
+            # Insert command.
+            result = sock_info.command(self.__database.name,
+                                       command,
+                                       codec_options=self.codec_options,
+                                       check_keys=check_keys)
+            _check_write_command_response([(0, result)])
+        else:
+            # Legacy OP_INSERT.
+            self._legacy_write(
+                sock_info, 'insert', command, acknowledged, op_id,
+                message.insert, self.__full_name, [doc], check_keys,
+                acknowledged, concern, False, self.codec_options)
+        return doc.get('_id')
+
+    def _insert(self, sock_info, docs, ordered=True, check_keys=True,
+                manipulate=False, write_concern=None, op_id=None):
         """Internal insert helper."""
-        return_one = False
         if isinstance(docs, collections.MutableMapping):
-            return_one = True
-            docs = [docs]
+            return self._insert_one(
+                sock_info, docs, ordered,
+                check_keys, manipulate, write_concern, op_id)
 
         ids = []
 
@@ -415,29 +495,29 @@ class Collection(common.BaseObject):
                     yield doc
 
         concern = (write_concern or self.write_concern).document
-        safe = concern.get("w") != 0
-
-        if sock_info.max_wire_version > 1 and safe:
-            # Insert command.
-            command = SON([('insert', self.name),
-                           ('ordered', ordered)])
-
-            if concern:
-                command['writeConcern'] = concern
+        acknowledged = concern.get("w") != 0
 
+        command = SON([('insert', self.name),
+                       ('ordered', ordered)])
+        if concern:
+            command['writeConcern'] = concern
+        if op_id is None:
+            op_id = message._randint()
+        bwc = message._BulkWriteContext(
+            self.database.name, command, sock_info, op_id,
+            self.database.client._event_listeners)
+        if sock_info.max_wire_version > 1 and acknowledged:
+            # Batched insert command.
             results = message._do_batched_write_command(
-                self.database.name + ".$cmd", _INSERT, command,
-                gen(), check_keys, self.codec_options, sock_info)
+                self.database.name + ".$cmd", message._INSERT, command,
+                gen(), check_keys, self.codec_options, bwc)
             _check_write_command_response(results)
         else:
             # Legacy batched OP_INSERT.
             message._do_batched_insert(self.__full_name, gen(), check_keys,
-                                       safe, concern, not ordered,
-                                       self.codec_options, sock_info)
-        if return_one:
-            return ids[0]
-        else:
-            return ids
+                                       acknowledged, concern, not ordered,
+                                       self.codec_options, bwc)
+        return ids
 
     def insert_one(self, document):
         """Insert a single document.
@@ -458,6 +538,8 @@ class Collection(common.BaseObject):
         :Returns:
           - An instance of :class:`~pymongo.results.InsertOneResult`.
 
+        .. seealso:: :ref:`writes-and-ids`
+
         .. versionadded:: 3.0
         """
         common.validate_is_mutable_mapping("document", document)
@@ -468,7 +550,7 @@ class Collection(common.BaseObject):
                                    self.write_concern.acknowledged)
 
     def insert_many(self, documents, ordered=True):
-        """Insert a list of documents.
+        """Insert an iterable of documents.
 
           >>> db.test.count()
           0
@@ -479,7 +561,7 @@ class Collection(common.BaseObject):
           2
 
         :Parameters:
-          - `documents`: A list of documents to insert.
+          - `documents`: A iterable of documents to insert.
           - `ordered` (optional): If ``True`` (the default) documents will be
             inserted on the server serially, in the order provided. If an error
             occurs all remaining inserts are aborted. If ``False``, documents
@@ -489,9 +571,11 @@ class Collection(common.BaseObject):
         :Returns:
           An instance of :class:`~pymongo.results.InsertManyResult`.
 
+        .. seealso:: :ref:`writes-and-ids`
+
         .. versionadded:: 3.0
         """
-        if not isinstance(documents, list) or not documents:
+        if not isinstance(documents, collections.Iterable) or not documents:
             raise TypeError("documents must be a non-empty list")
         inserted_ids = []
         def gen():
@@ -501,40 +585,39 @@ class Collection(common.BaseObject):
                 if "_id" not in document:
                     document["_id"] = ObjectId()
                 inserted_ids.append(document["_id"])
-                yield (_INSERT, document)
+                yield (message._INSERT, document)
 
         blk = _Bulk(self, ordered)
         blk.ops = [doc for doc in gen()]
         blk.execute(self.write_concern.document)
         return InsertManyResult(inserted_ids, self.write_concern.acknowledged)
 
-    def _update(self, sock_info, filter, document, upsert=False,
+    def _update(self, sock_info, criteria, document, upsert=False,
                 check_keys=True, multi=False, manipulate=False,
-                write_concern=None):
+                write_concern=None, op_id=None, ordered=True):
         """Internal update / replace helper."""
-        common.validate_is_mapping("filter", filter)
         common.validate_boolean("upsert", upsert)
         if manipulate:
             document = self.__database._fix_incoming(document, self)
-
         concern = (write_concern or self.write_concern).document
-        safe = concern.get("w") != 0
-
-        if sock_info.max_wire_version > 1 and safe:
+        acknowledged = concern.get("w") != 0
+        command = SON([('update', self.name),
+                       ('ordered', ordered),
+                       ('updates', [SON([('q', criteria),
+                                         ('u', document),
+                                         ('multi', multi),
+                                         ('upsert', upsert)])])])
+        if concern:
+            command['writeConcern'] = concern
+        if sock_info.max_wire_version > 1 and acknowledged:
             # Update command.
-            command = SON([('update', self.name)])
-            if concern:
-                command['writeConcern'] = concern
 
-            docs = [SON([('q', filter), ('u', document),
-                         ('multi', multi), ('upsert', upsert)])]
-
-            results = message._do_batched_write_command(
-                self.database.name + '.$cmd', _UPDATE, command,
-                docs, check_keys, self.codec_options, sock_info)
-            _check_write_command_response(results)
-
-            _, result = results[0]
+            # The command result has to be published for APM unmodified
+            # so we make a shallow copy here before adding updatedExisting.
+            result = sock_info.command(self.__database.name,
+                                       command,
+                                       codec_options=self.codec_options).copy()
+            _check_write_command_response([(0, result)])
             # Add the updatedExisting field for compatibility.
             if result.get('n') and 'upserted' not in result:
                 result['updatedExisting'] = True
@@ -546,13 +629,12 @@ class Collection(common.BaseObject):
                     result['upserted'] = result['upserted'][0]['_id']
 
             return result
-
         else:
             # Legacy OP_UPDATE.
-            request_id, msg, max_size = message.update(
-                self.__full_name, upsert, multi, filter, document, safe,
-                concern, check_keys, self.codec_options)
-            return sock_info.legacy_write(request_id, msg, max_size, safe)
+            return self._legacy_write(
+                sock_info, 'update', command, acknowledged, op_id,
+                message.update, self.__full_name, upsert, multi, criteria,
+                document, acknowledged, concern, check_keys, self.codec_options)
 
     def replace_one(self, filter, replacement, upsert=False):
         """Replace a single document matching the filter.
@@ -595,6 +677,7 @@ class Collection(common.BaseObject):
 
         .. versionadded:: 3.0
         """
+        common.validate_is_mapping("filter", filter)
         common.validate_ok_for_replace(replacement)
         with self._socket_for_writes() as sock_info:
             result = self._update(sock_info, filter, replacement, upsert)
@@ -632,6 +715,7 @@ class Collection(common.BaseObject):
 
         .. versionadded:: 3.0
         """
+        common.validate_is_mapping("filter", filter)
         common.validate_ok_for_update(update)
         with self._socket_for_writes() as sock_info:
             result = self._update(sock_info, filter, update,
@@ -670,6 +754,7 @@ class Collection(common.BaseObject):
 
         .. versionadded:: 3.0
         """
+        common.validate_is_mapping("filter", filter)
         common.validate_ok_for_update(update)
         with self._socket_for_writes() as sock_info:
             result = self._update(sock_info, filter, update, upsert,
@@ -686,34 +771,33 @@ class Collection(common.BaseObject):
         """
         self.__database.drop_collection(self.__name)
 
-    def _delete(self, sock_info, filter, multi, write_concern=None):
+    def _delete(
+            self, sock_info, criteria, multi,
+            write_concern=None, op_id=None, ordered=True):
         """Internal delete helper."""
-        common.validate_is_mapping("filter", filter)
+        common.validate_is_mapping("filter", criteria)
         concern = (write_concern or self.write_concern).document
-        safe = concern.get("w") != 0
+        acknowledged = concern.get("w") != 0
+        command = SON([('delete', self.name),
+                       ('ordered', ordered),
+                       ('deletes', [SON([('q', criteria),
+                                         ('limit', int(not multi))])])])
+        if concern:
+            command['writeConcern'] = concern
 
-        if sock_info.max_wire_version > 1 and safe:
+        if sock_info.max_wire_version > 1 and acknowledged:
             # Delete command.
-            command = SON([('delete', self.name)])
-            if concern:
-                command['writeConcern'] = concern
-
-            docs = [SON([('q', filter), ('limit', int(not multi))])]
-
-            results = message._do_batched_write_command(
-                self.database.name + '.$cmd', _DELETE, command,
-                docs, False, self.codec_options, sock_info)
-            _check_write_command_response(results)
-
-            _, result = results[0]
+            result = sock_info.command(self.__database.name,
+                                       command,
+                                       codec_options=self.codec_options)
+            _check_write_command_response([(0, result)])
             return result
-
         else:
             # Legacy OP_DELETE.
-            request_id, msg, max_size = message.delete(
-                self.__full_name, filter, safe, concern,
-                self.codec_options, int(not multi))
-            return sock_info.legacy_write(request_id, msg, max_size, safe)
+            return self._legacy_write(
+                sock_info, 'delete', command, acknowledged, op_id,
+                message.delete, self.__full_name, criteria, acknowledged,
+                concern, self.codec_options, int(not multi))
 
     def delete_one(self, filter):
         """Delete a single document matching the filter.
@@ -1267,8 +1351,8 @@ class Collection(common.BaseObject):
         codec_options = CodecOptions(SON)
         coll = self.with_options(codec_options)
         with self._socket_for_primary_reads() as (sock_info, slave_ok):
+            cmd = SON([("listIndexes", self.__name), ("cursor", {})])
             if sock_info.max_wire_version > 2:
-                cmd = SON([("listIndexes", self.__name), ("cursor", {})])
                 cursor = self._command(sock_info, cmd, slave_ok,
                                        ReadPreference.PRIMARY,
                                        codec_options)["cursor"]
@@ -1277,7 +1361,8 @@ class Collection(common.BaseObject):
                 namespace = _UJOIN % (self.__database.name, "system.indexes")
                 res = helpers._first_batch(
                     sock_info, namespace, {"ns": self.__full_name},
-                    0, slave_ok, codec_options, ReadPreference.PRIMARY)
+                    0, slave_ok, codec_options, ReadPreference.PRIMARY,
+                    cmd, self.database.client._event_listeners)
                 data = res["data"]
                 cursor = {
                     "id": res["cursor_id"],
@@ -1333,19 +1418,20 @@ class Collection(common.BaseObject):
             cursor = self.__database._list_collections(sock_info,
                                                        slave_ok,
                                                        criteria)
-            result = None
-            for doc in cursor:
-                result = doc
-                break
 
-            if not result:
-                return {}
+        result = None
+        for doc in cursor:
+            result = doc
+            break
 
-            options = result.get("options", {})
-            if "create" in options:
-                del options["create"]
+        if not result:
+            return {}
 
-            return options
+        options = result.get("options", {})
+        if "create" in options:
+            del options["create"]
+
+        return options
 
     def aggregate(self, pipeline, **kwargs):
         """Perform an aggregation using the aggregation framework on this
@@ -1602,7 +1688,7 @@ class Collection(common.BaseObject):
         .. versionchanged:: 2.2
            Removed deprecated arguments: merge_output and reduce_output
 
-        .. _map reduce command: http://www.mongodb.org/display/DOCS/MapReduce
+        .. _map reduce command: http://docs.mongodb.org/manual/reference/command/mapReduce/
 
         .. mongodoc:: mapreduce
         """
