@@ -20,6 +20,7 @@ MongoDB.
    application developers.
 """
 
+import datetime
 import random
 import struct
 
@@ -57,6 +58,11 @@ _OP_MAP = {
 }
 
 
+def _randint():
+    """Generate a pseudo random 32 bit integer."""
+    return random.randint(MIN_INT32, MAX_INT32)
+
+
 def _maybe_add_read_preference(spec, read_preference):
     """Add $readPreference to spec when appropriate."""
     mode = read_preference.mode
@@ -75,14 +81,129 @@ def _maybe_add_read_preference(spec, read_preference):
     return spec
 
 
+def _convert_exception(exception):
+    """Convert an Exception into a failure document for publishing."""
+    return {'errmsg': str(exception),
+            'errtype': exception.__class__.__name__}
+
+
+def _convert_write_result(operation, command, result):
+    """Convert a legacy write result to write commmand format."""
+
+    # Based on _merge_legacy from bulk.py
+    affected = result.get("n", 0)
+    res = {"ok": 1, "n": affected}
+    errmsg = result.get("errmsg", result.get("err", ""))
+    if errmsg:
+        # The write was successful on at least the primary so don't return.
+        if result.get("wtimeout"):
+            res["writeConcernError"] = {"errmsg": errmsg,
+                                        "code": 64,
+                                        "errInfo": {"wtimeout": True}}
+        else:
+            # The write failed.
+            error = {"index": 0,
+                     "code": result.get("code", 8),
+                     "errmsg": errmsg}
+            if "errInfo" in result:
+                error["errInfo"] = result["errInfo"]
+            res["writeErrors"] = [error]
+            return res
+    if operation == "insert":
+        # GLE result for insert is always 0 in most MongoDB versions.
+        res["n"] = len(command['documents'])
+    elif operation == "update":
+        if "upserted" in result:
+            res["upserted"] = [{"index": 0, "_id": result["upserted"]}]
+        # Versions of MongoDB before 2.6 don't return the _id for an
+        # upsert if _id is not an ObjectId.
+        elif result.get("updatedExisting") is False and affected == 1:
+            # If _id is in both the update document *and* the query spec
+            # the update document _id takes precedence.
+            update = command['updates'][0]
+            _id = update["u"].get("_id", update["q"].get("_id"))
+            res["upserted"] = [{"index": 0, "_id": _id}]
+    return res
+
+
+_OPTIONS = SON([
+    ('tailable', 2),
+    ('oplogReplay', 8),
+    ('noCursorTimeout', 16),
+    ('awaitData', 32),
+    ('allowPartialResults', 128)])
+
+
+_MODIFIERS = SON([
+    ('$query', 'filter'),
+    ('$orderby', 'sort'),
+    ('$hint', 'hint'),
+    ('$comment', 'comment'),
+    ('$maxScan', 'maxScan'),
+    ('$maxTimeMS', 'maxTimeMS'),
+    ('$max', 'max'),
+    ('$min', 'min'),
+    ('$returnKey', 'returnKey'),
+    ('$showRecordId', 'showRecordId'),
+    ('$showDiskLoc', 'showRecordId'),  # <= MongoDb 3.0
+    ('$snapshot', 'snapshot')])
+
+
+def _gen_explain_command(
+        coll, spec, projection, skip, limit, batch_size, options):
+    """Generate an explain command document."""
+    cmd = _gen_find_command(
+        coll, spec, projection, skip, limit, batch_size, options)
+    return SON([('explain', cmd)])
+
+
+def _gen_find_command(coll, spec, projection, skip, limit, batch_size, options):
+    """Generate a find command document."""
+    cmd = SON([('find', coll)])
+    if '$query' in spec:
+        cmd.update([(_MODIFIERS[key], val)
+                    for key, val in spec.items() if key in _MODIFIERS])
+    else:
+        cmd['filter'] = spec
+
+    if projection:
+        cmd['projection'] = projection
+    if skip:
+        cmd['skip'] = skip
+    if limit:
+        cmd['limit'] = abs(limit)
+        if limit < 0:
+            cmd['singleBatch'] = True
+    if batch_size:
+        cmd['batchSize'] = batch_size
+
+    if options:
+        cmd.update([(opt, True)
+                    for opt, val in _OPTIONS.items()
+                    if options & val])
+    return cmd
+
+
+def _gen_get_more_command(cursor_id, coll, batch_size, max_time_ms):
+    """Generate a getMore command document."""
+    cmd = SON([('getMore', cursor_id),
+               ('collection', coll)])
+    if batch_size:
+        cmd['batchSize'] = batch_size
+    if max_time_ms:
+        cmd['maxTimeMS'] = max_time_ms
+    return cmd
+
+
 class _Query(object):
     """A query operation."""
 
-    __slots__ = ('flags', 'ns', 'ntoskip', 'ntoreturn',
-                 'spec', 'fields', 'codec_options', 'read_preference')
+    __slots__ = ('flags', 'ns', 'ntoskip', 'ntoreturn', 'spec', 'fields',
+                 'codec_options', 'read_preference', 'limit', 'batch_size',
+                 'name')
 
-    def __init__(self, flags, ns, ntoskip, ntoreturn,
-                 spec, fields, codec_options, read_preference):
+    def __init__(self, flags, ns, ntoskip, ntoreturn, spec, fields,
+                 codec_options, read_preference, limit, batch_size):
         self.flags = flags
         self.ns = ns
         self.ntoskip = ntoskip
@@ -91,37 +212,90 @@ class _Query(object):
         self.fields = fields
         self.codec_options = codec_options
         self.read_preference = read_preference
+        self.limit = limit
+        self.batch_size = batch_size
+        self.name = 'find'
+
+    def as_command(self):
+        """Return a find command document for this query.
+
+        Should be called *after* get_message.
+        """
+        dbn, coll = self.ns.split('.', 1)
+        if '$explain' in self.spec:
+            self.name = 'explain'
+            return _gen_explain_command(
+                coll, self.spec, self.fields, self.ntoskip,
+                self.limit, self.batch_size, self.flags), dbn
+        return _gen_find_command(coll, self.spec, self.fields, self.ntoskip,
+                                 self.limit, self.batch_size, self.flags), dbn
 
     def get_message(self, set_slave_ok, is_mongos):
         """Get a query message, possibly setting the slaveOk bit."""
         if is_mongos:
-            spec = _maybe_add_read_preference(self.spec, self.read_preference)
-        else:
-            spec = self.spec
+            self.spec = _maybe_add_read_preference(self.spec,
+                                                   self.read_preference)
         if set_slave_ok:
             # Set the slaveOk bit.
             flags = self.flags | 4
         else:
             flags = self.flags
-        return query(flags, self.ns, self.ntoskip,
-                     self.ntoreturn, spec, self.fields, self.codec_options)
+        return query(flags, self.ns, self.ntoskip, self.ntoreturn,
+                     self.spec, self.fields, self.codec_options)
 
 
 class _GetMore(object):
     """A getmore operation."""
 
-    __slots__ = ('ns', 'ntoreturn', 'cursor_id')
+    __slots__ = ('ns', 'ntoreturn', 'cursor_id', 'max_time_ms')
 
-    def __init__(self, ns, ntoreturn, cursor_id):
+    name = 'getMore'
+
+    def __init__(self, ns, ntoreturn, cursor_id, max_time_ms=None):
         self.ns = ns
         self.ntoreturn = ntoreturn
         self.cursor_id = cursor_id
+        self.max_time_ms = max_time_ms
+
+    def as_command(self):
+        """Return a getMore command document for this query."""
+        dbn, coll = self.ns.split('.', 1)
+        return _gen_get_more_command(
+            self.cursor_id, coll, self.ntoreturn, self.max_time_ms), dbn
 
     def get_message(self, dummy0, dummy1):
         """Get a getmore message."""
         return get_more(self.ns, self.ntoreturn, self.cursor_id)
 
 
+class _CursorAddress(tuple):
+    """The server address (host, port) of a cursor, with namespace property."""
+
+    def __new__(cls, address, namespace):
+        self = tuple.__new__(cls, address)
+        self.__namespace = namespace
+        return self
+
+    @property
+    def namespace(self):
+        """The namespace this cursor."""
+        return self.__namespace
+
+    def __hash__(self):
+        # Two _CursorAddress instances with different namespaces
+        # must not hash the same.
+        return (self + (self.__namespace,)).__hash__()
+
+    def __eq__(self, other):
+        if isinstance(other, _CursorAddress):
+            return (tuple(self) == tuple(other)
+                    and self.namespace == other.namespace)
+        return NotImplemented
+
+    def __ne__(self, other):
+        return not self == other
+
+
 def __last_error(namespace, args):
     """Data to send to do a lastError.
     """
@@ -137,7 +311,7 @@ def __pack_message(operation, data):
 
     Returns the resultant message string.
     """
-    request_id = random.randint(MIN_INT32, MAX_INT32)
+    request_id = _randint()
     message = struct.pack("<i", 16 + len(data))
     message += struct.pack("<i", request_id)
     message += _ZERO_32  # responseTo
@@ -147,11 +321,7 @@ def __pack_message(operation, data):
 
 def insert(collection_name, docs, check_keys,
            safe, last_error_args, continue_on_error, opts):
-    """Get an **insert** message.
-
-    Used by the Bulk API to insert into pre-2.6 servers. Collection.insert
-    uses _do_batched_insert.
-    """
+    """Get an **insert** message."""
     options = 0
     if continue_on_error:
         options += 1
@@ -203,14 +373,14 @@ if _use_c:
 
 
 def query(options, collection_name, num_to_skip,
-          num_to_return, query, field_selector, opts):
+          num_to_return, query, field_selector, opts, check_keys=False):
     """Get a **query** message.
     """
     data = struct.pack("<I", options)
     data += bson._make_c_string(collection_name)
     data += struct.pack("<i", num_to_skip)
     data += struct.pack("<i", num_to_return)
-    encoded = bson.BSON.encode(query, False, opts)
+    encoded = bson.BSON.encode(query, check_keys, opts)
     data += encoded
     max_bson_size = len(encoded)
     if field_selector is not None:
@@ -269,9 +439,123 @@ def kill_cursors(cursor_ids):
     return __pack_message(2007, data)
 
 
+_FIELD_MAP = {
+    'insert': 'documents',
+    'update': 'updates',
+    'delete': 'deletes'
+}
+
+
+class _BulkWriteContext(object):
+    """A wrapper around SocketInfo for use with write splitting functions."""
+
+    __slots__ = ('db_name', 'command', 'sock_info', 'op_id',
+                 'name', 'field', 'publish', 'start_time', 'listeners')
+
+    def __init__(
+            self, database_name, command, sock_info, operation_id, listeners):
+        self.db_name = database_name
+        self.command = command
+        self.sock_info = sock_info
+        self.op_id = operation_id
+        self.listeners = listeners
+        self.publish = listeners.enabled_for_commands
+        self.name = next(iter(command))
+        self.field = _FIELD_MAP[self.name]
+        self.start_time = datetime.datetime.now() if self.publish else None
+
+    @property
+    def max_bson_size(self):
+        """A proxy for SockInfo.max_bson_size."""
+        return self.sock_info.max_bson_size
+
+    @property
+    def max_message_size(self):
+        """A proxy for SockInfo.max_message_size."""
+        return self.sock_info.max_message_size
+
+    @property
+    def max_write_batch_size(self):
+        """A proxy for SockInfo.max_write_batch_size."""
+        return self.sock_info.max_write_batch_size
+
+    def legacy_write(self, request_id, msg, max_doc_size, acknowledged, docs):
+        """A proxy for SocketInfo.legacy_write that handles event publishing.
+        """
+        if self.publish:
+            duration = datetime.datetime.now() - self.start_time
+            cmd = self._start(request_id, docs)
+            start = datetime.datetime.now()
+        try:
+            result = self.sock_info.legacy_write(
+                request_id, msg, max_doc_size, acknowledged)
+            if self.publish:
+                duration = (datetime.datetime.now() - start) + duration
+                if result is not None:
+                    reply = _convert_write_result(self.name, cmd, result)
+                else:
+                    # Comply with APM spec.
+                    reply = {'ok': 1}
+                self._succeed(request_id, reply, duration)
+        except OperationFailure as exc:
+            if self.publish:
+                duration = (datetime.datetime.now() - start) + duration
+                self._fail(
+                    request_id,
+                    _convert_write_result(
+                        self.name, cmd, exc.details),
+                    duration)
+            raise
+        finally:
+            self.start_time = datetime.datetime.now()
+        return result
+
+    def write_command(self, request_id, msg, docs):
+        """A proxy for SocketInfo.write_command that handles event publishing.
+        """
+        if self.publish:
+            duration = datetime.datetime.now() - self.start_time
+            self._start(request_id, docs)
+            start = datetime.datetime.now()
+        try:
+            reply = self.sock_info.write_command(request_id, msg)
+            if self.publish:
+                duration = (datetime.datetime.now() - start) + duration
+                self._succeed(request_id, reply, duration)
+        except OperationFailure as exc:
+            if self.publish:
+                duration = (datetime.datetime.now() - start) + duration
+                self._fail(request_id, exc.details, duration)
+            raise
+        finally:
+            self.start_time = datetime.datetime.now()
+        return reply
+
+    def _start(self, request_id, docs):
+        """Publish a CommandStartedEvent."""
+        cmd = self.command.copy()
+        cmd[self.field] = docs
+        self.listeners.publish_command_start(
+            cmd, self.db_name,
+            request_id, self.sock_info.address, self.op_id)
+        return cmd
+
+    def _succeed(self, request_id, reply, duration):
+        """Publish a CommandSucceededEvent."""
+        self.listeners.publish_command_success(
+            duration, reply, self.name,
+            request_id, self.sock_info.address, self.op_id)
+
+    def _fail(self, request_id, failure, duration):
+        """Publish a CommandFailedEvent."""
+        self.listeners.publish_command_failure(
+            duration, failure, self.name,
+            request_id, self.sock_info.address, self.op_id)
+
+
 def _do_batched_insert(collection_name, docs, check_keys,
                        safe, last_error_args, continue_on_error, opts,
-                       sock_info):
+                       ctx):
     """Insert `docs` using multiple batches.
     """
     def _insert_message(insert_message, send_safe):
@@ -291,14 +575,16 @@ def _do_batched_insert(collection_name, docs, check_keys,
     data.write(bson._make_c_string(collection_name))
     message_length = begin_loc = data.tell()
     has_docs = False
+    to_send = []
     for doc in docs:
         encoded = bson.BSON.encode(doc, check_keys, opts)
         encoded_length = len(encoded)
-        too_large = (encoded_length > sock_info.max_bson_size)
+        too_large = (encoded_length > ctx.max_bson_size)
 
         message_length += encoded_length
-        if message_length < sock_info.max_message_size and not too_large:
+        if message_length < ctx.max_message_size and not too_large:
             data.write(encoded)
+            to_send.append(doc)
             has_docs = True
             continue
 
@@ -306,7 +592,7 @@ def _do_batched_insert(collection_name, docs, check_keys,
             # We have enough data, send this message.
             try:
                 request_id, msg = _insert_message(data.getvalue(), send_safe)
-                sock_info.legacy_write(request_id, msg, 0, send_safe)
+                ctx.legacy_write(request_id, msg, 0, send_safe, to_send)
             # Exception type could be OperationFailure or a subtype
             # (e.g. DuplicateKeyError)
             except OperationFailure as exc:
@@ -326,18 +612,19 @@ def _do_batched_insert(collection_name, docs, check_keys,
                                    " - the connected server supports"
                                    " BSON document sizes up to %d"
                                    " bytes." %
-                                   (encoded_length, sock_info.max_bson_size))
+                                   (encoded_length, ctx.max_bson_size))
 
         message_length = begin_loc + encoded_length
         data.seek(begin_loc)
         data.truncate()
         data.write(encoded)
+        to_send = [doc]
 
     if not has_docs:
         raise InvalidOperation("cannot do an empty bulk insert")
 
     request_id, msg = _insert_message(data.getvalue(), safe)
-    sock_info.legacy_write(request_id, msg, 0, safe)
+    ctx.legacy_write(request_id, msg, 0, safe, to_send)
 
     # Re-raise any exception stored due to continue_on_error
     if last_error is not None:
@@ -347,11 +634,11 @@ if _use_c:
 
 
 def _do_batched_write_command(namespace, operation, command,
-                              docs, check_keys, opts, sock_info):
+                              docs, check_keys, opts, ctx):
     """Execute a batch of insert, update, or delete commands.
     """
-    max_bson_size = sock_info.max_bson_size
-    max_write_batch_size = sock_info.max_write_batch_size
+    max_bson_size = ctx.max_bson_size
+    max_write_batch_size = ctx.max_write_batch_size
     # Max BSON object size + 16k - 2 bytes for ending NUL bytes.
     # Server guarantees there is enough room: SERVER-10643.
     max_cmd_size = max_bson_size + 16382
@@ -390,6 +677,8 @@ def _do_batched_write_command(namespace, operation, command,
     # Where to write list document length
     list_start = buf.tell() - 4
 
+    to_send = []
+
     def send_message():
         """Finalize and send the current OP_QUERY message.
         """
@@ -403,11 +692,11 @@ def _do_batched_write_command(namespace, operation, command,
         buf.seek(command_start)
         buf.write(struct.pack('<i', length - command_start))
         buf.seek(4)
-        request_id = random.randint(MIN_INT32, MAX_INT32)
+        request_id = _randint()
         buf.write(struct.pack('<i', request_id))
         buf.seek(0)
         buf.write(struct.pack('<i', length))
-        return sock_info.write_command(request_id, buf.getvalue())
+        return ctx.write_command(request_id, buf.getvalue(), to_send)
 
     # If there are multiple batches we'll
     # merge results in the caller.
@@ -446,10 +735,12 @@ def _do_batched_write_command(namespace, operation, command,
             idx_offset += idx
             idx = 0
             key = b'0'
+            to_send = []
         buf.write(_BSONOBJ)
         buf.write(key)
         buf.write(_ZERO_8)
         buf.write(value)
+        to_send.append(doc)
         idx += 1
 
     if not has_docs:
