@@ -14,17 +14,31 @@
 
 """Internal network layer helper methods."""
 
+import datetime
 import select
 import struct
 
+_HAS_POLL = True
+_poller = None
+_EVENT_MASK = 0
+try:
+    from select import poll
+    _poller = poll()
+    _EVENT_MASK = (select.POLLIN | select.POLLPRI | select.POLLERR |
+                   select.POLLHUP | select.POLLNVAL)
+except ImportError:
+    _HAS_POLL = False
+
 from pymongo import helpers, message
-from pymongo.errors import AutoReconnect
+from pymongo.errors import AutoReconnect, NotMasterError, OperationFailure
 
 _UNPACK_INT = struct.Struct("<i").unpack
 
 
-def command(sock, dbname, spec, slave_ok, is_mongos, read_preference,
-            codec_options, check=True, allowable_errors=None):
+def command(sock, dbname, spec, slave_ok, is_mongos,
+            read_preference, codec_options, check=True,
+            allowable_errors=None, address=None,
+            check_keys=False, listeners=None):
     """Execute a command over the socket, or raise socket.error.
 
     :Parameters:
@@ -37,21 +51,55 @@ def command(sock, dbname, spec, slave_ok, is_mongos, read_preference,
       - `codec_options`: a CodecOptions instance
       - `check`: raise OperationFailure if there are errors
       - `allowable_errors`: errors to ignore if `check` is True
+      - `address`: the (host, port) of `sock`
+      - `check_keys`: if True, check `spec` for invalid keys
+      - `listeners`: An instance of :class:`~pymongo.monitoring.EventListeners`
     """
+    name = next(iter(spec))
     ns = dbname + '.$cmd'
     flags = 4 if slave_ok else 0
+    # Publish the original command document.
+    orig = spec
     if is_mongos:
         spec = message._maybe_add_read_preference(spec, read_preference)
+
+    publish = listeners is not None and listeners.enabled_for_commands
+    if publish:
+        start = datetime.datetime.now()
+
     request_id, msg, _ = message.query(flags, ns, 0, -1, spec,
-                                       None, codec_options)
-    sock.sendall(msg)
-    response = receive_message(sock, 1, request_id)
-    unpacked = helpers._unpack_response(response, codec_options=codec_options)
-    response_doc = unpacked['data'][0]
-    msg = "command %s on namespace %s failed: %%s" % (
-        repr(spec).replace("%", "%%"), ns)
-    if check:
-        helpers._check_command_response(response_doc, msg, allowable_errors)
+                                       None, codec_options, check_keys)
+
+    if publish:
+        encoding_duration = datetime.datetime.now() - start
+        listeners.publish_command_start(orig, dbname, request_id, address)
+        start = datetime.datetime.now()
+
+    try:
+        sock.sendall(msg)
+        response = receive_message(sock, 1, request_id)
+        unpacked = helpers._unpack_response(
+            response, codec_options=codec_options)
+
+        response_doc = unpacked['data'][0]
+        if check:
+            msg = "command %s on namespace %s failed: %%s" % (
+                repr(spec).replace("%", "%%"), ns)
+            helpers._check_command_response(response_doc, msg, allowable_errors)
+    except Exception as exc:
+        if publish:
+            duration = (datetime.datetime.now() - start) + encoding_duration
+            if isinstance(exc, (NotMasterError, OperationFailure)):
+                failure = exc.details
+            else:
+                failure = message._convert_exception(exc)
+            listeners.publish_command_failure(
+                duration, failure, name, request_id, address)
+        raise
+    if publish:
+        duration = (datetime.datetime.now() - start) + encoding_duration
+        listeners.publish_command_success(
+            duration, response_doc, name, request_id, address)
     return response_doc
 
 
@@ -93,7 +141,12 @@ def socket_closed(sock):
     """Return True if we know socket has been closed, False otherwise.
     """
     try:
-        rd, _, _ = select.select([sock], [], [], 0)
+        if _HAS_POLL:
+            _poller.register(sock, _EVENT_MASK)
+            rd = _poller.poll(0)
+            _poller.unregister(sock)
+        else:
+            rd, _, _ = select.select([sock], [], [], 0)
     # Any exception here is equally bad (select.error, ValueError, etc.).
     except:
         return True
