@@ -244,22 +244,35 @@ class TestClient(IntegrationTest):
         self.assertFalse(client_context.rs_or_standalone_client != c)
 
     def test_host_w_port(self):
-        with self.assertRaises(AutoReconnect):
+        with self.assertRaises(ValueError):
             connected(MongoClient("%s:1234567" % host, connectTimeoutMS=1,
                                   serverSelectionTimeoutMS=10))
 
     def test_repr(self):
-        # Making host a str avoids the 'u' prefix in Python 2, so the repr is
-        # the same in Python 2 and 3.
-        self.assertEqual(repr(MongoClient(str(host), port)),
-                         "MongoClient('%s', %d)" % (host, port))
+        # Used to test 'eval' below.
+        import bson
+
+        client = MongoClient(
+            'mongodb://localhost:27017,localhost:27018/?replicaSet=replset'
+            '&connectTimeoutMS=12345',
+            connect=False, document_class=SON)
+
+        the_repr = repr(client)
+        self.assertIn('MongoClient(host=', the_repr)
+        self.assertIn(
+            "document_class=bson.son.SON, "
+            "tz_aware=False, "
+            "connect=False, ",
+            the_repr)
+        self.assertIn("connecttimeoutms='12345'", the_repr)
+        self.assertIn("replicaset=", the_repr)
+
+        self.assertEqual(eval(the_repr), client)
 
     @client_context.require_replica_set
     def test_repr_replica_set(self):
-        # Like MongoClient(["localhost:27017", "localhost:27018"]).
-        self.assertIn("MongoClient([", repr(self.client))
-        for node in client_context.nodes:
-            self.assertIn("%s:%d" % node, repr(self.client))
+        self.assertIn("MongoClient(host=[", repr(self.client))
+        self.assertIn(pair, repr(self.client))
 
     def test_getters(self):
         self.assertEqual(client_context.client.address, (host, port))
@@ -387,13 +400,14 @@ class TestClient(IntegrationTest):
             raise SkipTest("UNIX-sockets are not supported on this system")
 
         mongodb_socket = '/tmp/mongodb-27017.sock'
+        encoded_socket = '%2Ftmp%2Fmongodb-27017.sock'
         if not os.access(mongodb_socket, os.R_OK):
             raise SkipTest("Socket file is not accessible")
 
         if client_context.auth_enabled:
-            uri = "mongodb://%s:%s@%s" % (db_user, db_pwd, mongodb_socket)
+            uri = "mongodb://%s:%s@%s" % (db_user, db_pwd, encoded_socket)
         else:
-            uri = "mongodb://%s" % mongodb_socket
+            uri = "mongodb://%s" % encoded_socket
 
         # Confirm we can do operations via the socket.
         client = MongoClient(uri)
@@ -404,7 +418,7 @@ class TestClient(IntegrationTest):
         # Confirm it fails with a missing socket.
         self.assertRaises(
             ConnectionFailure,
-            connected, MongoClient("mongodb:///tmp/non-existent.sock",
+            connected, MongoClient("mongodb://%2Ftmp%2Fnon-existent.sock",
                                    serverSelectionTimeoutMS=100))
 
     def test_fork(self):
@@ -469,7 +483,6 @@ class TestClient(IntegrationTest):
         self.assertEqual(SON, c.codec_options.document_class)
         self.assertTrue(isinstance(db.test.find_one(), SON))
 
-
     def test_timeouts(self):
         client = rs_or_single_client(connectTimeoutMS=10500)
         self.assertEqual(10.5, get_pool(client).opts.connect_timeout)
@@ -534,12 +547,14 @@ class TestClient(IntegrationTest):
             'mongodb://localhost/?serverSelectionTimeoutMS=0', connect=False)
         self.assertAlmostEqual(0, client.server_selection_timeout)
 
-        self.assertRaises(ValueError, MongoClient,
-                          'mongodb://localhost/?serverSelectionTimeoutMS=-1',
-                          connect=False)
-        self.assertRaises(ValueError, MongoClient,
-                          'mongodb://localhost/?serverSelectionTimeoutMS=',
-                          connect=False)
+        # Test invalid timeout in URI ignored and set to default.
+        client = MongoClient(
+            'mongodb://localhost/?serverSelectionTimeoutMS=-1', connect=False)
+        self.assertAlmostEqual(30, client.server_selection_timeout)
+
+        client = MongoClient(
+            'mongodb://localhost/?serverSelectionTimeoutMS=', connect=False)
+        self.assertAlmostEqual(30, client.server_selection_timeout)
 
     def test_waitQueueTimeoutMS(self):
         client = rs_or_single_client(waitQueueTimeoutMS=2000)
