@@ -27,7 +27,9 @@ import gridfs
 from bson.binary import Binary
 from bson.py3compat import u, StringIO, string_type
 from pymongo.mongo_client import MongoClient
-from pymongo.errors import ConfigurationError, ConnectionFailure
+from pymongo.errors import (ConfigurationError,
+                            ConnectionFailure,
+                            ServerSelectionTimeoutError)
 from pymongo.read_preferences import ReadPreference
 from gridfs.errors import CorruptGridFile, FileExists, NoFile
 from test.test_replica_set_client import TestReplicaSetClientBase
@@ -172,24 +174,21 @@ class TestGridfs(IntegrationTest):
         finally:
             self.fs.delete(files_id)
 
-    def test_delete_ensures_index(self):
+    def test_put_ensures_index(self):
         # setUp has dropped collections.
         names = self.db.collection_names()
         self.assertFalse([name for name in names if name.startswith('fs')])
 
         chunks = self.db.fs.chunks
+        files = self.db.fs.files
+        self.fs.put(b"junk")
 
-        self.fs.delete(file_id=1)
-
-        # delete() has ensured an index on (files_id, n).
-        # index_information() is like:
-        # {
-        #     '_id_': {'key': [('_id', 1)]},
-        #     'files_id_1_n_1': {'key': [('files_id', 1), ('n', 1)]}
-        # }
         self.assertTrue(any(
             info.get('key') == [('files_id', 1), ('n', 1)]
             for info in chunks.index_information().values()))
+        self.assertTrue(any(
+            info.get('key') == [('filename', 1), ('uploadDate', 1)]
+            for info in files.index_information().values()))
 
     def test_alt_collection(self):
         oid = self.alt.put(b"hello world")
@@ -244,7 +243,7 @@ class TestGridfs(IntegrationTest):
         # Should have created 100 versions of 'test' file
         self.assertEqual(
             100,
-            self.db.fs.files.find({'filename':'test'}).count()
+            self.db.fs.files.find({'filename': 'test'}).count()
         )
 
     def test_get_last_version(self):
@@ -402,11 +401,11 @@ class TestGridfs(IntegrationTest):
                              serverSelectionTimeoutMS=10)
         db = client.db
         gfs = gridfs.GridFS(db)
-        self.assertRaises(ConnectionFailure, gfs.list)
+        self.assertRaises(ServerSelectionTimeoutError, gfs.list)
 
         fs = gridfs.GridFS(db)
-        f = fs.new_file()  # Still no connection.
-        self.assertRaises(ConnectionFailure, f.close)
+        f = fs.new_file()
+        self.assertRaises(ServerSelectionTimeoutError, f.close)
 
     def test_gridfs_find(self):
         self.fs.put(b"test2", filename="two")
@@ -416,7 +415,7 @@ class TestGridfs(IntegrationTest):
         self.fs.put(b"test1", filename="one")
         time.sleep(0.01)
         self.fs.put(b"test2++", filename="two")
-        self.assertEqual(3, self.fs.find({"filename":"two"}).count())
+        self.assertEqual(3, self.fs.find({"filename": "two"}).count())
         self.assertEqual(4, self.fs.find().count())
         cursor = self.fs.find(
             no_cursor_timeout=False).sort("uploadDate", -1).skip(1).limit(2)
@@ -465,7 +464,7 @@ class TestGridfsReplicaSet(TestReplicaSetClientBase):
 
     def test_gridfs_replica_set(self):
         rsc = rs_client(
-            w=self.w, wtimeout=5000,
+            w=self.w,
             read_preference=ReadPreference.SECONDARY)
 
         fs = gridfs.GridFS(rsc.pymongo_test)
